/*
 * This file is part of HIN (Http INterceptor), a Chromium extension
 * for use in privacy enhancement, web development and testing.
 *
 * options.js -- Javascript for options form.  This script should be
 * loaded only by options.html edit form and nowhere else.
 *
 * Copyright 2013, 2014 by John K. Hinsdale.  See LICENSE.html for terms of use.
 */

// Allowed keys in the actions list
var GLOBAL__actions_keys = [ "name", "enabled", "description", "action" ];

// Allowed operations in an action, along with earliest event handler
// in which it can be executed.
var GLOBAL__action_info = {
    // onBeforeRequest
    "cancel": { args: 0, event: "onBeforeRequest" },
    "redirect": { args: 1, event: "onBeforeRequest" },
    "alert": { args: 1, event: "onBeforeRequest" },
    "debug": { args: 0, event: "onBeforeRequest" },
    // onBeforeSendHeaders
    "set_request_header": { args: [2, 4], event: "onBeforeSendHeaders" },
    "drop_request_header": { args: 1, event: "onBeforeSendHeaders" },
    // onHeadersReceived
    "set_response_header": { args: [2, 4], event: "onHeadersReceived" },
    "drop_response_header": { args: 1, event: "onHeadersReceived" },
    // onAuthRequired
    "authorize": { args: 2, event: "onAuthRequired" }
};

// Disable check on non-intercptable headers until we can figure out what the complete list
// For example, Host: does seem to be able to be trapped and replaced(!)
/*
// Un-trappable request headers.  Warn if try to set/drop them.
var GLOBAL__untrappable_request_headers = [
    "Authorization", "Cache-Control", "Connection", "Content-Length",
    "Host", "If-Modified-Since", "If-None-Match", "If-Range", "Partial-Data",
    "Pragma", "Proxy-Authorization", "Proxy-Connection", "Transfer-Encoding"
];
*/

// Allowed keys in the rules list
var GLOBAL__rules_keys = [ "name", "enabled", "description", "match", "action" ];
// String operators in match expression
var GLOBAL__string_ops = [ "contains", "starts", "ends", "equals", "matches" ];

// List of processing phases, in order they occur.  This will map to
// an integer (index into the array) indicating how early a rule may
// be matched, or action list executed, in its entirety.  Match
// operators and action events will map to these names, and in turn,
// to points on the request/response "timeline" that can then be used
// to decide when to execute actions.
// The "Blk?" means the handler can (and will) block, i.e. run synchronously.
// The "*" means the event may occur multiple times in the request.
var GLOBAL__events = [
    //                        Blk? * Match-on        Action(s)
    // === Request ===        ---- - --------------  ------------------------------------------
    "onBeforeRequest",     // Blk    method, url     cancel, redirect
    "onBeforeSendHeaders", // Blk    request_header  cancel, (redirect?), set/drop_req_header
    "onSendHeaders",
    // === Response ===
    "onHeadersReceived",   // Blk  * response_header redirect, set/drop_resp_header
    "onAuthRequired",      // Blk  *                 authorize
    "onBeforeRedirect",
    "onResponseStarted",
    "onErrorOccurred",
    "onCompleted"
];

// Options export/import file name
GLOBAL__export_filename = "hin-options.json";

/*
 * Compile options for use in request checking.  Check duplicates,
 * syntax (ops and args), validate and pre-compile regex expressions,
 * remove disabled items, build auxiliary lookup structures, etc.
 *
 * Return hash of compiles rules and actions, or errors list.
 */
function compile_options(opts) {

    // Clean out old stuff first
    remove_compiled_info(opts);

    // Compile actions
    var c_actions = compile_actions(opts.actions);

    // Build actions map for rules compilation
    var action_map = {};
    for (var i=0; i < opts.actions.length; i++ ) {
        var act = opts.actions[i];
        if ( ("name" in act) && act.name.length > 0 )
            action_map[act.name] = act;
    }

    // Compile rules
    var c_rules = compile_rules(opts.rules, action_map);

    // Gather errors
    var result = {};
    if ( c_actions.errors || c_rules.errors ) {
        result.errors = [];
        if ( c_actions.errors )
            result.errors = result.errors.concat(c_actions.errors);
        if ( c_rules.errors )
            result.errors = result.errors.concat(c_rules.errors);
        console.log("COMPILE ERRORS:\n" + JSON.stringify(result.errors, null, 2));
        return result;
    }

    // Get enabled actions, by name
    var actions_by_name = {};
    for ( var k in action_map )
        if ( action_map[k].enabled )
            actions_by_name[k] = action_map[k];
    opts.actions_by_name = actions_by_name;

    // For each rule, build a map of event handler to list of action
    // components to run when it matches.  In each rule, each event handler
    // for the rule maps to a list of pairs (action name, operation), where
    // operations are segregated out from the action's compiled_action list
    // (already filtered for disabled actions).
    //    <event> -> [ [Action Name], [ <op>, [<args>] ] ]
    for (var i=0; i < opts.rules.length; i++) {
        var rule = opts.rules[i];
        if ( !rule.enabled )
            continue;
        rule.event_actions = {};
        for (var j=0; j < rule.compiled_action.length; j++) {
            var actname = rule.compiled_action[j];
            var act = actions_by_name[actname];
            var ops = act.compiled_action;
            for (k=0; k < ops.length; k++) {
                var op_args = ops[k];
                var op = op_args[0];
                if ( ! (op in GLOBAL__action_info) )
                    return { errors: ["Internal error: cannot find event for operation '" + op + "' in action '" + actname + "'"] };
                var event = GLOBAL__action_info[op].event;
                if ( ! (event in rule.event_actions) )
                    rule.event_actions[event] = [];
                var pair = [actname, op_args];
                rule.event_actions[event].push(pair);
            }
        }
    }

    return { opts: opts };
}

// Compile rules, checking against defined actions
function compile_rules(rules, action_map) {
    var compiled_rules = [];
    var seen = {};
    var errors = [];
    for (var i=0; i < rules.length; i++ ) {
        var rule = rules[i];
        // Check rule names given are unique.  Unnamed rules will be
        // trapped by compile_rule() below.
        if ( "name" in rule ) {
            if ( rule.name in seen ) {
                errors.push("Duplicate rule name '" + rule.name + "'");
                continue;
            }
            else
                seen[rule.name] = true;
        }
        var c_rule = compile_rule(rule, action_map);
        if ( c_rule.errors ) {
            errors = errors.concat(c_rule.errors);
            continue;
        }
        // Skip valid, but disabled rules
        if ( !rule.enabled )
            continue;
        compiled_rules.push(c_rule.rule);
    }
    if ( errors.length )
        return { errors: errors };
    return { rules: compiled_rules };
}

// Compile rule.  Return hash with either "errors" or "actions"
function compile_rule(rule, action_map) {
    var errors = [];
    var rule_name = "<not named>";
    if ( "name" in rule )
        rule_name = rule.name;
    else
        errors.push("Rule entry has no 'name' field");

    // Check for garbage keys
    var valid_keys = array_to_hash(GLOBAL__rules_keys, true);
    var any_invalid = false;
    for (var k in rule)
        if ( !(k in valid_keys) ) {
            errors.push("Rule '" + rule_name + "' has unknown field '" + k + "' = '" + rule[k] + "'");
            any_invalid = true;
        }
    if ( any_invalid )
        errors.push("NOTE: Valid rule fields are: " + GLOBAL__rules_keys.join(", "));
        
    // Check for required key(s)
    if ( ! ("enabled" in rule) )
        errors.push("Rule '" + rule_name + "' has no 'enabled' field");

    // Compile the match expression
    var compiled_match = [];
    if ( "match" in rule ) {
        var match = rule.match;
        var m_result = compile_match(match, "In rule '" + rule_name + "' - ");
        if ( m_result.errors )
            errors = errors.concat(m_result.errors);
        else
            compiled_match = m_result.match;
    }
    else
        errors.push("Rule '" + rule_name + "' entry has no 'match' field");

    // Check action names and build list of actual actions executed by rule, organized
    // by handler, and stripping out disabled actions
    var compiled_action = [];
    if ( "action" in rule ) {
        var action = rule.action;
        // Convert string to single-element array
        if ( (typeof action) == "string" )
            action = [ rule.action ];
        // Check the action list: that each exists, and build a
        // compile_action list of enabled actions' names.  This will be used later
        // to segregate the enabled actions' operations by event type.
        var any_invalid = false;
        for (var i=0; i < action.length; i++)
            if ( action[i] in action_map ) {
                if ( action_map[action[i]].enabled )
                    compiled_action.push(action[i]);
            }
            else {
                any_invalid = true;
                errors.push("Rule '" + rule_name + "' refers to unknown action '" + action[i] + "'");
            }
        if ( any_invalid ) {
            var valid_actions = Object.keys(action_map);
            valid_actions.sort();
            errors.push("NOTE: Defined action names are:\n    " + valid_actions.join("\n    "));
        }
    }
    else
        errors.push("Rule '" + rule_name + "' entry has no 'action' field");
    
    // All done checking everything, return complete set of errors
    if ( errors.length > 0 )
        return { errors: errors };
    
    // console.log("Rule '" + rule.name + "' has compiled match:\n" + JSON.stringify(compiled_match, null, 2));
    rule.compiled_match = compiled_match;
    rule.compiled_action = compiled_action;
    return { rule: rule };
}


// Compile match expression
function compile_match(match, err_prefix) {
    // See if it's a scalar (boolean)
    if ( (typeof match) == "boolean" )
        return { match: match };
    // Otherwise it should be an array expression
    if ( ! is_array(match) )
        return { errors: [err_prefix + "Match expression '" + match + "' is of type '" + (typeof match) + "', not true/false or array"] };
    if ( match.length == 0 )
        return { errors: [err_prefix + "Match expression cannot be empty array"] };

    var errors = [];
    var compiled_match = [];
    // Handle prefix logic operators
    var op = match[0];
    if ( (typeof op) != "string" )
        return { errors: [err_prefix + "First item in match '" + op + "' not a string"] };
    var op_lc = op.toLowerCase();
    
    if ( op_lc == "and" || op_lc == "or" || op_lc == "not" ) {
        if ( (op_lc == "and" || op_lc == "or") && match.length < 2 )
            return { errors: [err_prefix + "Logical '" + op + "' must have at least one argument"] };
        if ( op == "not" && match.length != 2 )
             return { errors: [err_prefix + "Logical '" + op + "' must have at exactly one argument, not " + (match.length - 1)] };
        compiled_match.push(op);
        for (var i=1; i < match.length; i++) {
            var c_result = compile_match(match[i], err_prefix);
            if ( c_result.errors )
                errors = errors.concat(c_result.errors);
            else
                compiled_match.push(c_result.match);
        }
    }
    /* Handle match on field (url, method) ... at the moment we are only matching on method and URL
     (not on request or response headers), so the only ops we need to allow are the string ops.
     Numeric options (e.g. for Content-Length) and "is present" (e.g. for a header) can come later.
     For simplicity, all string operations are done case-insensitively. */
    else if ( op_lc == "url" || op_lc == "method" ) {
        // Check we have a string op and one arg, that arg either being a string or
        // or string array
        if ( match.length != 3 )
            return { errors: [err_prefix + "Match on '" + op + "' must be followed by an operator (contains, etc.) and argument"] };
        var string_op = match[1];
        if ( (typeof string_op) != "string" )
            return { errors: [err_prefix + "Match on '" + op + "': operator '" + string_op + "' not a string"] };
        var string_op_lc = string_op.toLowerCase();
        var valid_string_ops = array_to_hash(GLOBAL__string_ops);
        if ( ! (string_op_lc in valid_string_ops) )
            return { errors: [err_prefix + "Match on '" + op + "': operator '" + string_op +
                              "' unknown.  Valid match operators are: " + GLOBAL__string_ops.join(", ")] };

        // Handle argument to string op, either a single string, or array or strings
        // Normalize into list of strings, always
        var op_arg = match[2];
        if ( (typeof op_arg) == "string" )
            op_arg = [ match[2] ];
        else if ( ! is_array(op_arg) )
            return { errors: [err_prefix + "Match on '" + op + "': operator '" + string_op +
                              "' should take a string or list of strings, not '" + op_arg + "'"] };
        // Also check each arg in the list is a string, and that regexes are valid
        for (var i=0; i < op_arg.length; i++) {
            if ( (typeof op_arg[i]) != "string" )
                return { errors: [err_prefix + "Match on '" + op + "': operator '" + string_op +
                                  "' item in arg list should be a string, not '" + op_arg[i] + "'"] };
            else if ( string_op == "matches" ) {
                var re_info = get_regex(op_arg[i]);
                if ( ! re_info.valid )
                    return { errors: [err_prefix + "Invalid regular expression arg '" + op_arg[i] + "':\n" + re_info.error] };
            }
        }

        // All OK, add items to compiled expression, field/op normalized to lower case
        // and args to list
        compiled_match.push(op_lc);
        compiled_match.push(string_op_lc);
        compiled_match.push(op_arg);
    }
    else
        // First item not a known field
        return { errors: [err_prefix + "First item of match '" + op + "' not a field (url, method)"] };
        
    // Return error, or compiled result
    if ( errors.length )
        return { errors: errors };
    return { match: compiled_match };
}

// Compile actions.  Return object with error array "errors",
// or enabled action list "actions"
function compile_actions(actions) {
    var seen = {};
    var errors = [];
    for (var i=0; i < actions.length; i++ ) {
        var action = actions[i];
        // Check action names given are unique.  Unnamed actions will be
        // trapped by compile_action() below.
        if ( "name" in action ) {
            if ( action.name in seen ) {
                errors.push("Duplicate action name '" + action.name + "'");
                continue;
            }
            else
                seen[action.name] = true;
        }
        var c_action = compile_action(action);
        if ( c_action.errors ) {
            errors = errors.concat(c_action.errors);
            continue;
        }
        // Skip valid, but disabled actions
        if ( !action.enabled )
            action.compiled_action = [];
        else
            action.compiled_action = c_action.action;
    }
    if ( errors.length )
        return { errors: errors };
    return { actions: actions };
}

// Compile action.  Return a hash with either "errors" or "action"
// set.  For "action", always returns a list of entries, each of which
// is [ "op", "arg", ... ]
function compile_action(action) {
    var errors = [];
    var action_name = "<not named>";
    if ( "name" in action )
        action_name = action.name;
    else
        errors.push("Action entry has no 'name' field");

    // Check for garbage keys
    var valid_keys = array_to_hash(GLOBAL__actions_keys, true);
    var any_invalid = false;
    for (var k in action)
        if ( !(k in valid_keys) ) {
            errors.push("Action '" + action_name + "' has unknown field '" + k + "' = '" + action[k] + "'");
            any_invalid = true;
        }
    if ( any_invalid )
        errors.push("Valid action fields are: " + GLOBAL__actions_keys.join(", "));
        
    // Check for required key(s)
    if ( ! ("enabled" in action) )
        errors.push("Action '" + action_name + "' has no 'enabled' field");

    // Compile the action proper
    var c_result = {};
    if ( "action" in action ) {
        var act = action.action;
        c_result = compile_action_op(act, "In action '" + action_name + "' - ");
        if ( c_result.errors )
            errors = errors.concat(c_result.errors);
    }
    else 
        errors.push("Action '" + action_name + "' entry has no 'action' field");
    if ( errors.length )
        return { errors: errors };
    return { action: c_result.compiled_action };
}

// Compile an action, which can be one of:
//    (1) single string unary op, e.g. "cancel"
//    (2) single operation of form [ "op", "arg1", ... ]
//    (3) list of items of type (1) or type (2)
// Return result with key "errors" or "compiled_action"
// Arg err_prefix is for error messages
function compile_action_op(act, err_prefix) {
    var compiled_action = [], errors = [];
    // See if action is no-arg op like "cancel"
    var op, args;
    var is_list = false;
    if ( (typeof act) == "string" ) {
        op = act;
        args = [];
    }
    else {
        // Should be an array, either [ "op" ... ] or action list [ ["op" ..], [] ]
        if ( ! is_array(act) )
            return { errors: [err_prefix + "Action '" + act + "' is of type '" + (typeof act) + "', not string or array"] };
        if ( act.length > 0 ) {
            // Check if first element is a list, then treat each item as an action
            if ( is_array(act[0]) ) {
                // Compile action list, collecting errors
                for (var i=0; i < act.length; i++) {
                    var c_action = compile_action_op(act[i], err_prefix);
                    if ( c_action.errors )
                        errors = errors.concat(c_action.errors);
                    else
                        compiled_action = compiled_action.concat(c_action.compiled_action);
                }
                is_list = true;
            }
            else {
                // Splice out op and args
                args = act.slice(0); // Make a copy
                op = args.shift();
            }
        }
        else
            // Empty list (allowed)
            is_list = true;
    }

    if ( ! is_list ) {
        // Validate op is valid and number of args is as expected
        var actinfo = GLOBAL__action_info;
        if ( op in actinfo ) {
            var expected = actinfo[op].args;
            // Check range or exact value
            if ( is_array(expected) ) {
                var minarg = expected[0];
                var maxarg = expected[1];
                if ( args.length < minarg )
                    errors.push(err_prefix + "Got " + args.length + " args to operator '" + op + "', need at least " + minarg);
                else if ( args.length > maxarg )
                    errors.push(err_prefix + "Got " + args.length + " args to operator '" + op + "', can have at most " + minarg);
            }
            else if ( args.length != (expected) ) {
                errors.push(err_prefix + "Got " + args.length + " args to operator '" + op + "', expected " + expected);
            }
        }
        else {
            var valid_ops = Object.keys(actinfo);
            valid_ops.sort();
            errors.push(err_prefix + "Action operator '" + op + "' invalid.  Valid operators are:\n    " + valid_ops.join("\n    "));
        }
        
        // Op-specific validations
        if ( (op == "drop_request_header" || op == "drop_response_header"
              || op == "set_request_header" || op == "set_response_header")
             && args.length > 0
             && ends_with(args[0], ":") )
            errors.push(err_prefix + "Header name '" + args[0] + "' should not end with a colon");
        if ( op == "set_request_header" || op == "set_response_header") {
            // Should be 0=header_name 1=value 2=add_if_absent 3=overwrite_existing
            if ( (args.length > 2 && (typeof args[2]) != "boolean")
                 || (args.length > 3 && (typeof args[3]) != "boolean") )
                errors.push(err_prefix + "Last 2 args of set_*_header should be booleans (true/false)");
            // Validate that new value is either a string to set, or a sequence of substitutions
            var newval = args[1];
            if ( (typeof newval) != "string" ) {
                if ( !is_array(newval) )
                    errors.push(err_prefix + "New value for set_*_header should be string, pair (old, new), or list of pairs, not '" + newval + "'");
                else if ( newval.length == 0 )
                    errors.push(err_prefix + "New value for set_*_header cannot be empty array");
                else {
                    // Convert single pair to one-element list
                    if ( (typeof newval[0]) == "string" )
                        args[1] = newval = [ newval ];
                    // Validate list of pairs
                    for (var nv=0; nv < newval.length; nv++) {
                        var pair = newval[nv];
                        if ( !is_array(pair) )
                            errors.push(err_prefix + "Value in new value list for set_*_header should be a pair (old, new), not '" + pair + "'");
                        else if ( pair.length != 2 )
                            errors.push(err_prefix + "Value in new value list for set_*_header should be a pair of length 2 (old, new), not of length " + pair.length + "'");
                        else if ( pair.length != 2 )
                            errors.push(err_prefix + "Value in new value list for set_*_header should be a pair of length 2 (old, new), not of length " + pair.length + "'");
                        else if ( (typeof pair[0]) != "string" || (typeof pair[1]) != "string" )
                            errors.push(err_prefix + "Values in pair (old, new) should be string in set_*_header operation, not '" + pair + "'");
                        // Check that pair[0] is a valid regex
                        var re_inf = get_regex(pair[0]);
                        if ( !re_inf.valid )
                            errors.push(err_prefix + "Invalid regular expression substitution in set_*_header operation:\n" + pair[0] + "\n" + re_inf.error);
                    }
                }
                // Also ensure that "add_if_absent" is false, since we can only do a substitution on existing value
                if ( args.length < 3 || args[2] )
                    errors.push(err_prefix + "In set_*header, when substituion pair(s) are given, following arg add_if_absent must be false");
            }
        }
        
        // Disable check on un-trappable headers until we can figure out what the complete list is,
        // both for request and response headers
        /*
        // Check if trying to set or drop request headers that cannot be touched
        if ( (op == "set_request_header" || op == "drop_request_header")
             && args.length > 0 ) {
            var untrap = array_to_hash(GLOBAL__untrappable_request_headers, true);
            if ( args[0] in untrap )
                errors.push(err_prefix + "Request header '" + args[0] + "' cannot be intercepted for set/drop");
        }
        */
        
        // Authorize args should be 0=username 1=passwd
        if ( op == "authorize" && args.length != 2 )
            errors.push(err_prefix + "Need 2 args to authorize: username, passwd");
        
        if ( ! errors.length ) {
            // Looks good, set compiled action to [op, args...]
            var to_push = [ op ];
            if ( args.length > 0 )
                to_push = to_push.concat(args);
            compiled_action.push(to_push);
        }
    }

    // Return error or compiled action
    if ( errors.length )
        return { errors: errors };
    return { compiled_action: compiled_action };
}
        
// Verify options and alert if all OK
function verify_options() {
    var v = check_options();
    if ( v.success )
        alert("Options are valid.");
}

// Check options.  Return result with success tag = true if all is OK
function check_options() {
    var opts = get_form_options();
    result = { success: false };
    if ( opts.error ) {
        var problem_obj;
        if ( !opts.actions_ok )
            problem_obj = "actions list";
        else if ( !opts.rules_ok )
            problem_obj = "rules list";
        alert("Cannot parse JSON in the " + problem_obj + ":\n\n" + opts.message);
        return result;
    };

    // Try compiling
    var c_opts = compile_options(opts);
    if ( c_opts.errors ) {
        var msg = "Error(s) occurred:\n";
        for (var i=0; i < c_opts.errors.length; i++)
            msg += "\n" + c_opts.errors[i] + "\n";
        alert(msg);
        return result;
    }
    
    // All looks OK ... Put validated/edited options back on the form
    // after verify.  This will reformat the JSOn
    set_form_options(opts);
    result.success = true;
    result.opts = opts;
    // console.log("Result of check:\n" + JSON.stringify(result, null, 4));
    return result;
}

// Get options from form, apply for use, and save to storage
function save_options() {
    var v_result = check_options();
    if ( !v_result.success )
        return;
    // Options look OK, save.
    var opts = v_result.opts;
    set_global_options(opts);
    b = chrome.extension.getBackgroundPage();
    var cleaned_opts = clean_options(opts);
    b.save_options_to_storage(
        cleaned_opts,
        function() {
            flash_status("Options saved.");
        }
    );
}

// Load options from storage, populate on form, and apply for use
function load_options() {
    chrome.storage.local.get(
        get_global_options_default(),
        function(dirty_opts) {
            // Just in case compiled info was stored persistently, clean it up
            // This is also a good place to convert data in deprecated form
            var opts = clean_options(dirty_opts);
            compile_options(opts);
            set_form_options(opts);
            set_global_options(opts);
            update_icon();
        }
    );
}

// Revert options
function revert_options() {
    load_options();
    flash_status("Options reverted to last saved values.");
}

// Reset options in storage to defaults, confirming, then populate form with reset values
function reset_options() {
    // Confirm
    if ( ! confirm("Really reset options?  All customizations will be lost!") ) {
        flash_status("Reset cancelled.");
        return;
    }
    default_opts = get_global_options_default();
    chrome.storage.local.set(
        default_opts,
        function() {
            // Update status to let user know options were saved.
            flash_status('Options reset.');
        }
    );
    // Now populate the form and apply
    compile_options(default_opts);
    set_form_options(default_opts);
    set_global_options(default_opts);
    update_icon();
}

// Get options from form
function get_form_options() {
    var enabled = !! document.getElementById('enabled').checked;
    var debug = !! document.getElementById('debug').checked;
    var cookies = !! document.getElementById('cookies').checked;
    var newtab_url = document.getElementById('newtab_url').value.trim();
    var debug_requests_count = document.getElementById('debug_requests_count').value.trim();
    var enable_icon = document.getElementById('enable_icon').checked;
    var actions_json = document.getElementById('actions').value;
    var rules_json = document.getElementById('rules').value;

    // Convert form JSON to JS data
    var actions, rules;
    var actions_ok  = false, rules_ok = false, count_ok = false;
    try {
        actions = JSON.parse(actions_json);
        actions_ok = true;
        rules = JSON.parse(rules_json);
        rules_ok = true;
        debug_requests_count = parseInt(debug_requests_count);
        count_ok = true;
    }
    catch (exc) {
        // Try to get a message from exc
        if ( !actions_ok || !rules_ok )
            return { error: true, message: "Invalid JSON: " + exc.toString(), exception: exc,
                     actions_ok: actions_ok, rules_ok: rules_ok, count_ok: count_ok };
        if ( !count_ok )
            return { error: true, message: "Invalid debug request count:  " + exc.toString(), exception: exc,
                     actions_ok: actions_ok, rules_ok: rules_ok, count_ok: count_ok };
    };
        
    // All OK
    return { enabled: enabled, debug: debug, cookies: cookies, enable_icon: enable_icon,
             newtab_url: newtab_url, debug_requests_count: debug_requests_count,
             actions: actions, rules: rules };
}

// Set options on to the form
function set_form_options(opts) {
    document.getElementById('enabled').checked = opts.enabled;
    document.getElementById('debug').checked = opts.debug;
    document.getElementById('cookies').checked = opts.cookies;
    document.getElementById('newtab_url').value = opts.newtab_url.trim();
    document.getElementById('debug_requests_count').value = opts.debug_requests_count;
    document.getElementById('enable_icon').checked = opts.enable_icon;
    
    // Convert actions and rules lists to JSON
    var indent = 2;
    document.getElementById('actions').value = JSON.stringify(opts.actions, GLOBAL__actions_keys, indent);
    document.getElementById('rules').value = JSON.stringify(opts.rules, GLOBAL__rules_keys, indent);

    // Set the incognito warning
    chrome.extension.isAllowedIncognitoAccess(
        function(allowed) {
            var incog_warn = "";
            if( ! allowed )
                incog_warn = "<p><b>Warning:</b>  Your browser is currently set not to allow HIN in 'Incognito' (private) mode.  "
                             + "This is probably not what you want.  To change this setting, "
                             + "click to go to the <a href=\"#\" id=\"extensions_link\">"
                             + "browser extensions</a> page and check the box labelled \"Allow in incognito\" (private mode) "
                             + "in the section for HIN.  This page will then close automatically, "
                             + "and you will need to select \"HIN Options\" again to get back here (and this "
                             + "warning should be gone).<p>";
            document.getElementById('incognito_warning').innerHTML = incog_warn;
            // We'll only have the link if we set it above ...
            if( ! allowed )
                document.getElementById('extensions_link').addEventListener('click', open_extensions);
        }
    );
}

// Get options from the background page
function get_global_options() {
    b = chrome.extension.getBackgroundPage();
    return b.GLOBAL__options;
}
// Put options on the background page (e.g., after save or revert)
function set_global_options(opts) {
    b = chrome.extension.getBackgroundPage();
    b.GLOBAL__options = opts;
    // Any time we set the options we need to flush out cached pages that may
    // have been handled under old rules
    chrome.webRequest.handlerBehaviorChanged();
    b.apply_options();

}
// Get options defaults from the background page
function get_global_options_default() {
    b = chrome.extension.getBackgroundPage();
    return b.GLOBAL__options_default;
}

// "Clean" the options by removing all info added by compilation
// Useful just for storing and after loading so that only user
// specified stuff is ever stored/retrieved
function clean_options(src) {
    // console.log("******************************************************\n");
    // console.log("BEFORE CLEAN:\n" + JSON.stringify(src, null, 2) + "\n\n");

    var result = deep_clone(src);
    remove_compiled_info(result);
    // console.log("******************************************************\n");
    // console.log("AFTER CLEAN:\n" + JSON.stringify(result, null, 2));
    // console.log("******************************************************\n");
    return result;
}

// Remove compiled info from options
function remove_compiled_info(opts) {
    var i;
    for (i=0; i < opts.rules.length; i++){
        delete opts.rules[i]["compiled_match"];    // Normalized match expression
        delete opts.rules[i]["compiled_action"];   // List of enabled actions
        delete opts.rules[i]["event_actions"];     // Map of events to operations across all enabled actions
    }
    for (i=0; i < opts.actions.length; i++){
        delete opts.actions[i]["compiled_action"]; // Normalized action
    }
}

// Open extensions page
function open_extensions() {
    chrome.windows.create({url: "chrome://extensions/?id=" + chrome.runtime.id, focused: true});
}

function flash_status(msg) {
    // Update status to let user know options were saved.
    var status = document.getElementById('status');
    status.textContent = msg;
    setTimeout(function() {
        status.textContent = '';
    }, 750);
}

// Set icon path depending on enabled checkbox to re-inforce
// the meaning of the two icons
function update_icon() {
    img = document.getElementById('options_icon');
    enabled = document.getElementById('enabled');
    var disabled_warning = "";
    if ( enabled.checked )
        img.src = "icons/Fighter_jet_active_icon_48.png";
    else {
        img.src = "icons/Fighter_jet_disabled_icon_48.png";
        disabled_warning = "<p><b>Warning:</b>  HIN is disabled globally, and will have no effect.  Check \"Enable HIN\" below to enable.";
    }
    document.getElementById('disabled_warning').innerHTML = disabled_warning;
}

// Warn when turning off the toolbar icon
function warn_on_disable_icon() {
    var enable_icon = document.getElementById('enable_icon');
    if ( ! enable_icon.checked ) {
        var msg = "NOTE: after the HIN toolbar icon is disabled, options can be edited via the browser extensions page, \"Settings ... Extensions,\" and choosing the \"Options\" link in the section for HIN.\n\nThis will be the only way to re-enable the toolbar icon after the present changes to options are saved.\n\nReally disable the extension toolbar icon?";
        if ( ! confirm(msg) )
            enable_icon.checked = true;
    }
}

// Export options
function export_options() {
    var v_result = check_options();
    if ( !v_result.success )
        return;
    if ( ! confirm("Save options to your Downloads folder under the name '" + GLOBAL__export_filename + "'?") )
        return;
    // Options look OK, save.
    var opts = v_result.opts;
    var cleaned_opts = clean_options(opts);
    // Also put in version
    var curopts = get_global_options();
    cleaned_opts.options_version = curopts.options_version;
    var a = document.createElement('a');
    var json = JSON.stringify(cleaned_opts, null, 2);
    a.setAttribute('href', 'data:text/plain;charset=utf-8,' + encodeURIComponent(json));
    a.setAttribute('download', GLOBAL__export_filename);
    a.click();
}

// Import options
function import_options(ev) {
    var fl = ev.target.files[0]; 
    if ( fl ) {
      var rdr = new FileReader();
      rdr.onload = function(e) { 
          var fn = fl.name;
	  var contents = e.target.result;
          // Try to parse JSON
          var opts;
          try {
              opts = JSON.parse(contents);
          }
          catch (exc) {
              // Try to get a message from exc
              alert("Error in file '" + fn + "':\n" + exc.toString());
              return;
          }
          // Have opts ... set the global options and remind that you still need to save
          var cur_opts = get_form_options();
          merge_options(cur_opts, opts);
          set_form_options(cur_opts);
          var v = check_options();
          if ( v.success )
              alert("Loaded options from '" + fn + "' ... You still need to save for them to be made permanent.");
          // Reset value of file input "import" - can only do this by resetting the outer form
          var imp_form = document.getElementById("import_form");
          imp_form.reset();
      }
      rdr.readAsText(fl);
    }
    else
        alert("Failed to import JSON file");
}

// Merge options on import, adding/overwriting destination.
// Don't worry too much about validating since that will happen
// prior to save.
function merge_options(dst, src) {
    for ( var k in src )
        if ( k == "rules" || k == "actions" )
            merge_options_list(dst[k], src[k]);
        else
            dst[k] = src[k];
}

// Merge list of rules or actions based on name
function merge_options_list(dst, src) {
    // Collect existing names by index
    var cur_names = {};
    var i;
    for (i=0; i < dst.length; i++)
        cur_names[dst[i].name] = i;
    var to_add = [];
    for (i=0; i < src.length; i++)
        if ( src[i].name in cur_names )
            dst[cur_names[src[i].name]] = src[i];
        else
            to_add.push(src[i]);
    for (i=0; i < to_add.length; i++)
        dst.push(to_add[i]);
}


// $Id: options.js,v 1.45 2014/12/28 04:18:11 hin Exp $
